# UpdateManager.py
# -*- Mode: Python; indent-tabs-mode: nil; tab-width: 4; coding: utf-8 -*-
import os
import subprocess
from apt import Cache
import apt_pkg
import sys
import time
import shutil
import dbus
import logging
import dbus.service
from apt.debfile import DebPackage
from dbus.mainloop.glib import DBusGMainLoop
DBusGMainLoop(set_as_default=True)

from .Core.MyCache import MyCache
from .UpdateManagerDbus import UpdateManagerDbusController,UPDATER_DBUS_INTERFACE,UPDATER_DBUS_PATH,UPDATER_DBUS_SERVICE
from .Core.UpdateList import UpdateList
from .backend import InstallBackend,get_backend
from .Core.Database import Sqlite3Server
from .Core.loop import mainloop
from .Core.DataAcquisition import UpdateMsgCollector

from gettext import gettext as _
from SystemUpdater.backend import DownloadBackend as downb
from SystemUpdater.Core.UpdaterConfigParser import UpgradeConfig
from SystemUpdater.Core.utils import get_broken_details,get_lis_from_cache
from SystemUpdater.Core.DpkgInstallProgress import LogInstallProgress

class UpdateManager():
    SELF_PKG_NAME = 'kylin-system-updater'
    GROUPS_PKG_NAME = 'kylin-update-desktop-config'
    INSTALL_ALONE_PROGRESS = "alone"
    SOURCE_PARTS_PATH = "/etc/apt/"
    SOURCE_FILE_NAME = "sources.list"

    def __init__(self,options):
        self.options = options
        self.cache = None
        self.update_list = None

        #表示后端的状态 注意：这个状态很重要 用不好整个后端代码就会卡住 
        self.is_working = InstallBackend.ACTION_DEFUALT_STATUS

        #建立dbus
        self.dbusController = self._setup_dbus()

        #连接数据库
        self.sqlite3_server = Sqlite3Server(self)

        #是否重启apt
        self.is_restart_aptdeamon = False

        #配置文件
        self.configs = UpgradeConfig("/var/lib/kylin-system-updater/")
        self.uuconfigs = UpgradeConfig(datadir = "/var/lib/unattended-upgrades/", name = "unattended-upgrades-policy.conf")

        #数据采集器
        self.collector = UpdateMsgCollector(self)

        #后台aptdaemon的语言环境
        self.aptdaemonLang = os.environ["LANGUAGE"]


    #检查是否需要重新启动aptdeamon 目前需要重启的有限速功能
    def restart_aptdeamon(self):
        if self.is_restart_aptdeamon == True:
            logging.warning("aptdeamon will be restart...")
            self.is_restart_aptdeamon = False
            self.dbusController.make_aptdeamon_restart()

    def run(self):
        """Start the daemon and listen for calls."""
        logging.info("Waiting for calls")
        try:
            mainloop.run()
        except KeyboardInterrupt:
            self.dbusController.Quit(None)

    #进行清空所有下载的文件
    def start_clean(self):
        try:
            clean_backend = get_backend(self, InstallBackend.ACTION_CLEAN)
            clean_backend.start()
        except Exception as e:
            logging.error(e)

    #进行修复破损的包的操作 apt install -f 
    def start_fix_broken(self):
        try:
            fix_backend = get_backend(self, InstallBackend.ACTION_FIX_BROKEN)
            fix_backend.start()
        except Exception as e:
            logging.error(e)
    
    #进行 dpkg --configure
    def start_fix_incomplete(self):
        try:
            fix_backend = get_backend(self, InstallBackend.ACTION_FIX_INCOMPLETE)
            fix_backend.start()
        except Exception as e:
            logging.error(e)

    #download only
    def download_backend(self, pkgs):
        try:
            download_backend = downb.DownloadBackend(self)
            download_backend.downloadpkgs(pkgs)
        except Exception as e:
            logging.error(str(e))

    #download only 2.0
    def start_download(self, pkgs):
        try:
            if self.options.no_update_source is False:
                self.dbusController.on_update_important_list()

            update_backend = get_backend(self, InstallBackend.ACTION_DOWNLOADONLY)
            update_backend.start(partial_upgrade_list = pkgs)
        except Exception as e:
            logging.error(str(e))

    #读取source.list 内容
    def _read_source_list(self):
        try:
            source_file_content = ''
            source_file_path = os.path.join(self.SOURCE_PARTS_PATH,self.SOURCE_FILE_NAME)
            with open(source_file_path) as source_file:
                source_file_content = source_file.read()
            return source_file_content
        except Exception as e:
            logging.error(str(e))
            return ''         

    #判断是否为光盘源
    def _is_disc_source(self):
        logging.info("Check: Whether to use CD-ROM source updates Successfully...")
        source_file_content = self._read_source_list()
        if "deb file:" in source_file_content:
            logging.info("Current use of CD-ROM source without network check...")
            return True
        else:
            return False

    #进行更新的操作
    def start_update(self):
        _success = True
        header = ''
        desc = ''

        try:
            #每次更新之前 重新读取配置文件 刷新参数
            self.configs.reReadConfigFiles()
            
            #光盘源的话
            if self._is_disc_source() ==  True:
                logging.info("Turn off NetworkCheck and CloseFiter and UpdateSource...")
                #关闭网络检查
                self.options.no_check_network = True
                #当判断为光盘源时 关闭源过滤
                self.options.close_filter = True
                #关闭 更新源模板
                self.options.no_update_source = True

            if self.options.no_check_network is False:
                #进行检查网络
                self.dbusController.UpdateDetectStatusChanged(5,_("Checking network connection"))
                _success,header,desc = self.dbusController.check_connectivity()
                if _success == False:
                    self.dbusController.DistUpdateDetectFinished(_success,[],header,desc)
                    return

            #从服务器端更新摸板
            if self.options.no_update_source is False:
                self.dbusController.UpdateDetectStatusChanged(10,_("Updating Source Template"))
                _success,header,desc = self.dbusController.on_update_important_list()
                if _success == False:
                    self.dbusController.DistUpdateDetectFinished(_success,[],header,desc)
                    return
            
            #不进行update
            if self.options.no_update:
                self.start_available()
                return

            #调用aptdeamon进行update
            update_backend = get_backend(self, InstallBackend.ACTION_UPDATE)
            update_backend.start()
        except Exception as e:
            logging.error(e)

    
    # 进行本地deb包安装的操作
    def start_deb_install(self, deb_path = "", _check_local_dep = False, _auto_satisfy = False):
        # _check_local_dep : 是否查询本地依赖
        # _auto_satisfy    : 是否通过网络下载依赖
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        try:
            deb_cache, ins = self._suit_install_mode(deb_path)
            if self._is_broken > 0 or not self.cacheSatisfy or self._need_downgrade:
            # 走 dpkg 安装流程，说明本地apt环境已经损坏,or dep not satisfied or need downgrade
                dep_satisfy, header, desc = self._deb_install(deb_cache, deb_path, _check_local_dep, ins)
                if dep_satisfy: 
                    self.dbusController.InstalldebFinished(True, header, desc)
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
            else:
            # apt 安装流程
                dep_satisfy, header, desc = self._attempt_depends(deb_cache, deb_path, _check_local_dep,_auto_satisfy, ins)
                if dep_satisfy:
                    try:
                        install_backend = get_backend(self, InstallBackend.ACTION_INSTALL_DEB)
                        install_backend.start_alone(partial_upgrade_list = deb_path, _is_install = _auto_satisfy)
                    except Exception as e:
                        logging.error(str(e))
                else:
                    self.dbusController.InstalldebFinished(False, header, desc)
        except Exception as e:
            logging.info(str(e))
            self.dbusController.InstalldebFinished(False, str(e), desc)

    #进行升级的操作
    def start_install(self,upgrade_mode,is_install = False,partial_upgrade_list = []):
        try:
            if is_install == True:
                install_backend = get_backend(self, InstallBackend.ACTION_INSTALL)
                install_backend.start(upgrade_mode,partial_upgrade_list)
            else:
                resolver_backend = get_backend(self, InstallBackend.ACTION_CHECK_RESOLVER)
                resolver_backend.start(upgrade_mode,partial_upgrade_list)
        except Exception as e:
            logging.error(e)

    #进行删除的操作
    def start_purge_pkgs(self,pkgs_list):
        try:
            # 是否有破损的包
            deb_cache = Cache()
            broken_count = deb_cache._depcache.broken_count
            if broken_count > 0:
                _success,header,desc = self._dpkg_purge_pkgs(pkgs_list)
                if _success == True:
                    logging.info(header)
                    self.dbusController.PurgePackagesFinished(_success,'',desc)
                else:
                    self.dbusController.PurgePackagesFinished(_success,header,desc)
            else:
                purge_backend = get_backend(self, InstallBackend.ACTION_REMOVE_PACKAGES)
                purge_backend.start(partial_upgrade_list = pkgs_list)
        except Exception as e:
            logging.error(e)

    def _dpkg_purge_pkgs(self,pkgs_list):
        success = False
        pkg_list_str = " ".join(pkgs_list)
        args = ["dpkg", "-r"]
        args.extend([pkg_list_str])

        p = subprocess.run(args, stdout=subprocess.PIPE,stderr=subprocess.STDOUT,text=True)
        success = p.returncode == 0
        return success,p.stdout,''

    #更新结束之后会调到此获取要升级的列表 and 更新cache 生成升级组列表JSON
    def start_available(self):
        _success,header,desc = self.refresh_cache()

        #特殊情况的处理 单独安装包需要直接退出 安装or卸载执行完毕后 还会调到start_available
        if _success == False and header == self.INSTALL_ALONE_PROGRESS:
            return
        else:
            upgrade_list = []
            if self.update_list != None:
                upgrade_list = self.update_list.local_upgrade_data.upgrade_groups + self.update_list.local_upgrade_data.single_pkgs
            #发送更新升级列表完成的标志
            self.dbusController.DistUpdateDetectFinished(_success,upgrade_list,header,desc)
        #检查安装完成之后需要重启吗
        self.is_reboot_required()
        #清空cache
        if self.cache != None and self.cache.get_changes():
            self.cache.clear()

    def refresh_cache(self):
        _success = True
        header = ''
        desc = ''
        try:
            #第一次进入 之后update不进入
            if self.cache is None:
                self.cache = MyCache(None)
            else:
                self.cache.open(None)
                self.cache._initDepCache()
        except AssertionError:
            header = _("Software index is broken")
            desc = _("It is impossible to install or remove any software. "
                    "Please use the package manager \"Synaptic\" or run "
                    "\"sudo apt-get install -f\" in a terminal to fix "
                    "this issue at first.")
            _success = False
            return _success,header,desc

        except SystemError as e:
            header = _("Could not initialize the package information")
            desc = _("An unresolvable problem occurred while "
                    "initializing the package information.\n\n"
                    "Please report this bug against the 'kylin-system-updater' "
                    "package and include the following error "
                    "message:\n") + str(e)
            _success = False
            return _success,header,desc

        #提高生成对象的位置
        self.update_list = UpdateList(self)

        #1、 检查出现安装过程异常重启 出现的话 进行异常修复
        if self.configs.getWithDefault("SystemStatus", "isabnormalreboot", False) == True:
            header = self.INSTALL_ALONE_PROGRESS
            self.configs.setValue("SystemStatus","isabnormalreboot",str(False),True)
            logging.warning("start fix Abnormal Reboot broken pkgs...")
            self.start_fix_broken()
            return False,header,desc
        
        #检查当前系统的状态 是否存在系统为破损状态 导致出现异常
        self._check_cache_broken(self.cache)

        #检查自己是否需要升级更新
        _success,header,desc = self._check_self_upgrade(self.cache)
        if _success == False:
            return _success,header,desc

        #检查更新分组配置表
        _success,header,desc = self._check_group_config(self.cache)
        if _success == False and self.configs.getWithDefault("ConfigPkgStatus", "mustexistconfigpkg", True) == True:
            return _success,header,desc

        try:    
            _success,header,desc = self.update_list.update(self.cache)
        except SystemError as e:
            pass
        return _success,header,desc

    def _check_self_upgrade(self,cache):
        _success = True
        header = ''
        desc = ''
        #2、 安装JSON分组配置文件包 安装完毕会重新调start_available --> here 安装失败就直接退出不会进行下面的操作 
        try:
            if self.SELF_PKG_NAME in cache:
                pkg_json = cache[self.SELF_PKG_NAME]
                #是否安装
                if pkg_json.is_installed and pkg_json.is_upgradable:
                    logging.info("Check: Self-Updater(%s) start upgrading From %s to %s...",self.SELF_PKG_NAME,\
                                    pkg_json.installed.source_version,pkg_json.candidate.source_version)
                    pkg_json.mark_upgrade()

                    self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                    #直接退出 进行安装程序
                    _success = False
                    header = self.INSTALL_ALONE_PROGRESS
                    return _success,header,desc
                else:
                    logging.info("Check: Self-Updater(%s:%s) No need to upgrade...",self.SELF_PKG_NAME,pkg_json.installed.source_version)
                    #直接退出 进行之后的动作
                    _success = True
                    return _success,header,desc
            else:
                #没有在cache中就认为不需要升级
                logging.error("Check: Self-Updater(%s) The upgrade package is not in Cache...",self.SELF_PKG_NAME)
                #直接退出 进行之后的动作
                _success = True
                return _success,header,desc                

        #FIXME: 错误处理未做 报告到控制面板
        except Exception as e:
            logging.error(e)
            _success = True
            return _success,header,desc

    def _check_group_config(self,cache):
        _success = True
        header = ''
        desc = ''
        #2、 安装JSON分组配置文件包 安装完毕会重新调start_available --> here 安装失败就直接退出不会进行下面的操作 
        try:
            if self.GROUPS_PKG_NAME in cache:
                pkg_json = cache[self.GROUPS_PKG_NAME]
                #是否安装
                if pkg_json.is_installed:
                    #是否可升级
                    if pkg_json.is_upgradable:
                        logging.info("Check: groups JSON ConfigPkgs(%s) start upgrading From %s to %s...",self.SELF_PKG_NAME,\
                                        pkg_json.installed.source_version,pkg_json.candidate.source_version)
                        pkg_json.mark_upgrade()

                        self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                        #直接退出
                        _success = False
                        header = self.INSTALL_ALONE_PROGRESS
                        return _success,header,desc
                    else:
                        logging.info("Check: ConfigPkgs(%s:%s) No need to upgrade...",self.GROUPS_PKG_NAME,pkg_json.installed.source_version)
                else:
                    logging.info("Check: groups JSON ConfigPkgs(%s) start new installing...",self.GROUPS_PKG_NAME)
                    pkg_json.mark_install()
                    self.start_install(InstallBackend.MODE_UPGRADE_SINGLE,True)
                    #直接退出
                    _success = False
                    header = self.INSTALL_ALONE_PROGRESS
                    return _success,header,desc
            else:
                header = _("Unable to get group configuration package, please contact administrator to solve")
                desc = _("groups JSON ConfigPkgs install failed")
                logging.error("Check: groups JSON ConfigPkgs(%s) install failed...",self.GROUPS_PKG_NAME)
                _success = False
                return _success,header,desc                

        #FIXME: 错误处理未做 报告到控制面板
        except Exception as e:
            logging.error(e)
            _success = False
            return _success,header,desc
        
        #FIXME:
        #3、 判断目录是JSON配置文件夹是否缺失 缺失后进行修复 卸载重新安装步骤
        # if not os.path.exists(INPUT_CONFIG_PATH):
        #     logging.info("groups JSON Config Path(%s) Missing and Trying to fix...",INPUT_CONFIG_PATH)
        #     #将软件包卸载 之后进行重新安装here --> purge --> start_available 进行判断是否安装未安装重新安装
        #     # self.start_install_alone(pkgs_purge = [self.GROUPS_PKG_NAME])
        #     #直接退出
        #     _success = False
        #     header = self.INSTALL_ALONE_PROGRESS
        #     return _success,header,desc

        return _success,header,desc

    def is_reboot_required(self):
        """If a reboot is required to get all changes into effect."""
        return os.path.exists(os.path.join(apt_pkg.config.find_dir("Dir"),
                                        "var/run/reboot-required"))

    def _check_cache_broken(self,cache):
        with cache.actiongroup():
            if cache.get_changes():
                cache.clear()
            #获取出现破损状态包的数量
            wouldDelete = cache._depcache.broken_count
            
            if wouldDelete > 0:
                #获取那些依赖关系不满足导致的问题
                logging.info(get_broken_details(cache))
            else:
                logging.info("Check: System Apt Cache for Broken Successfully...")
                return
            
            try:
                #计算依赖的解决方案
                cache.fix_broken()
            except SystemError as e:
                logging.error(e)
                return
            pkgs_install,pkgs_upgrade,pkgs_remove = get_lis_from_cache(cache)
            logging.warning("Fix broken Packages is need to install:%a and upgrade:%a and remove:%a",pkgs_install,pkgs_upgrade,pkgs_remove)

            if cache.get_changes():
                cache.clear()

    def _setup_dbus(self):
        # check if there is another g-a-i already and if not setup one
        # listening on dbus
        bus = dbus.SystemBus()
        try:
            bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                            bus,
                                            do_not_queue=True)
            logging.info("initiate dbus success ...")
            return  UpdateManagerDbusController(self, bus_name)
        except dbus.exceptions.NameExistsException:
            if self.options.replace is False:
                logging.critical("Another daemon is already running")
                sys.exit(1)
            logging.warning("Replacing already running daemon")
            
            retry_reboot_times = 0
            the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                            UPDATER_DBUS_PATH)
            the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                timeout=300)
            time.sleep(1)
            while True:
                retry_reboot_times = retry_reboot_times + 1
                #当重试次数超过5次时退出程序
                if retry_reboot_times > 5:
                    logging.critical("Reboot backend is Failed...")
                    sys.exit(1)
                try:
                    bus_name = dbus.service.BusName(UPDATER_DBUS_SERVICE,
                                                    bus,
                                                    do_not_queue=True)
                    logging.warning("Replacing already running daemon to Success...")
                    return  UpdateManagerDbusController(self, bus_name)
                except dbus.exceptions.NameExistsException:
                    the_other_guy = bus.get_object(UPDATER_DBUS_SERVICE,
                                                    UPDATER_DBUS_PATH)
                    the_other_guy.Quit(dbus_interface=UPDATER_DBUS_INTERFACE,
                                        timeout=300)
                    logging.error("Dbus has not withdrawn and retry reboot times:%d...",retry_reboot_times)
                    time.sleep(1)

    # 是否查找本地依赖
    def _attempt_depends(self,deb_cache, deb_path,_check_local_dep,_auto_satisfy, _install):  
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _local_satisfy = False
        error_string = ''
        error_desc = ''
        absolute_path, debname = os.path.split(deb_path)
        # 依赖不满足的情况
        if len(_install) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(_install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if depends.name in debfile and depends.candidate.version in debfile:
                                #FIXME:检查depends包的合法性
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(depends_pkg):
                        #本地依赖不满足 判断源下载
                        if _auto_satisfy:
                            _local_satisfy = True
                        elif not _auto_satisfy:
                            _local_satisfy = False
                            error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                            error_desc   = ",".join(_install)
                            logging.error(error_string+ error_desc)
                    else:
                        #将应用包与依赖包拷贝至archive目录安装
                        try:
                            if debname not in satisfy_list:
                                satisfy_list.append(debname)
                            for satisfy in satisfy_list:
                                shutil.copy(os.path.join(absolute_path,satisfy),"/var/cache/apt/archives/")
                                logging.info("move debpkg: %s",satisfy)
                            _local_satisfy = True
                        except Exception as e:
                            logging.info(str(e))
                return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and _auto_satisfy:
                _local_satisfy = True
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied will download")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
            elif not _check_local_dep and not _auto_satisfy:
                _local_satisfy = False
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied")
                    error_desc   = ",".join(_install)
                    logging.error(error_string+error_desc)
                    return _local_satisfy,error_string,error_desc
        # 依赖满足
        else:
            _local_satisfy = True
            error_string = ''
            error_desc=''
            return _local_satisfy,error_string,error_desc
    
    def _deb_install(self, deb_cache, deb_path, _check_local_dep, _install):
        depends_list = []
        depends_pkg = []
        satisfy_list = []
        depends_count = 0
        _status = False
        error_string = ''
        header = ''
        desc = ''
        absolute_path, debname = os.path.split(deb_path)
        # 依赖不满足的情况
        if len(_install) > 0 or len(self.noSatisfyList) > 0:
            if _check_local_dep: #查找本地
                # 需要查找本地依赖
                if len(self.noSatisfyList) > 0:
                    noSatisfyList = [ str(nS[self.noSatisfyList.index(nS)][0]+'_'+nS[self.noSatisfyList.index(nS)][1]) for nS in self.noSatisfyList]
                if len(_install) > 0:
                    for pkg in deb_cache:
                        if pkg.marked_install and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                        elif pkg.marked_upgrade and pkg.name != str(debname.split("_")[0]):
                            depends_pkg.append(pkg)
                if len(depends_pkg)>0 or len(noSatisfyList)>0: #查找本地deb包
                    depends_list = [debfile for debfile in os.listdir(absolute_path) if debfile.endswith(".deb")]
                    for depends in depends_pkg:
                        for debfile in depends_list:
                            if "%3a" in debfile:
                                debfile=debfile.replace("%3a",":")
                            if depends.name in debfile and depends.candidate.version in debfile:
                                depends_count += 1
                                satisfy_list.append(debfile)
                    for depends in noSatisfyList:
                        for debfile in depends_list:
                            if "%3a" in debfile:
                                debfile=debfile.replace("%3a",":")
                            if depends.split('_')[0] == debfile.split('_')[0] and depends.split('_')[1] == debfile.split('_')[1] and debfile not in satisfy_list:
                                depends_count += 1
                                satisfy_list.append(debfile)
                    if depends_count < len(noSatisfyList):
                        #本地依赖不满足
                        _status = False
                        error_string = str(debname.split("_")[0])+_("dependency is not satisfied")+":".join(noSatisfyList)
                        logging.error(error_string)
                        header = error_string
                        return False,header, desc
                    else:
                        #安装依赖包
                        if debname not in satisfy_list:
                            satisfy_list.append(debname)
                        for satisfy in satisfy_list:
                            try:
                                deb = DebPackage(os.path.join(absolute_path,satisfy))
                                iprogress = LogInstallProgress(deb_path)
                                ret = deb.install(install_progress=iprogress)
                            except Exception as e:
                                logging.error(e)
                                return False, str(e), desc
                        return True, header, desc
            elif not _check_local_dep:
                _status = False
                if _install:
                    error_string = str(debname.split("_")[0])+_("dependency is not satisfied")+":".join(_install)
                    logging.error(error_string)
                    header = error_string
                    return False,header, desc
        # 依赖满足
        else:
            try:
                deb = DebPackage(deb_path)
                iprogress = LogInstallProgress(deb_path)
                ret = deb.install(install_progress=iprogress)
                if ret != 0:
                   return False, iprogress.errormsg, desc
            except Exception as e:
                logging.error(e)
                return False, str(e), desc
            return True, header, desc

    def _suit_install_mode(self, deb_path):
        self._is_broken = False
        self.cacheSatisfy = False
        absolute_path, debname = os.path.split(deb_path)
        # 检查本地破损
        try:
            logging.info("Install package, open cache")
            deb_cache = Cache()
            logging.info("Install package, check broken")
            broken_count = deb_cache._depcache.broken_count
            deb = DebPackage(deb_path, deb_cache)
            deb.check(allow_downgrade=True)
            logging.info("Install package, required changes")
            (install, remove, unauth) = deb.required_changes    # need in cach
            if broken_count > 0:
                self._is_broken = True
            else :
                self._is_broken = False
        except KeyError:
            pass
        except Exception as e:
            logging.error(str(e))
        # 需要降级
        try:
            pkg = deb_cache[debname.split('_')[0]]
            # 判断":"在版本号中的情况
            if "%3a" in debname:
                debname=debname.replace("%3a",":")
            if pkg.is_installed and pkg.installed.source_version > debname.split('_')[1]:
                logging.info("Downgrade deb %s, from %s to %s.",\
                                debname.split('_')[0], pkg.installed.source_version,debname.split('_')[1])
                self._need_downgrade = True
            else:
                self._need_downgrade = False
        except Exception as e:
            logging.error(str(e))
            self._need_downgrade = False
        # 不满足的依赖列表
        depends = deb.depends
        self.noSatisfyList = self._gen_noSatisfyList(depends, deb_cache)
        _list = []
        # cache是否满足
        if len(install) == 0 and len(self.noSatisfyList) == 0:
            self.cacheSatisfy = True
        else:
            for ns in self.noSatisfyList:
                for or_group in ns:
                    for pkg in install:
                        if pkg == or_group[0]:
                            _list.append(ns)
            if len(_list) == len(self.noSatisfyList):
                self.cacheSatisfy = True
            else:
                self.cacheSatisfy = False
        logging.info("Cache satisfy is %r.",self.cacheSatisfy)
        return deb_cache, install
    
    def _gen_noSatisfyList(self, depends, deb_cache):
        _noSatisfyList = []
        _group_satify = False
        for or_group in depends:
            for deb in or_group:
                debname,ver,oper = deb
                if ":" in debname:
                    debname=debname.split(":")[0]
                try:
                    if debname not in deb_cache:
                        if  deb_cache.is_virtual_package(debname):
                            logging.warning("The cache has no package named '%s', this is a virtual package...", debname)
                            providers = deb_cache.get_providing_packages(debname)
                        # hear!
                        if len(providers) != 1:
                            continue
                        depname = providers[0].name
                        pkg = deb_cache[depname]
                        cand = deb_cache._depcache.get_candidate_ver(pkg._pkg)
                        if not cand:
                            _group_satify = False
                        if not apt_pkg.check_dep(cand.ver_str, oper, ver):
                            _group_satify = False
                        else:
                            _group_satify = True
                    else:
                        pkg = deb_cache[debname]
                    if pkg.installed and apt_pkg.check_dep(pkg.installed.source_version, oper, ver) and _group_satify == False:
                        _group_satify = True
                except Exception as e:
                    logging.error(str(e))
            if _group_satify == False and or_group not in _noSatisfyList:
                _noSatisfyList.append(or_group)
            _group_satify = False
        return _noSatisfyList
